#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This test shows a surrogate-assisted Bayesian calibration of a beam deflection
    model.

Author: Farid Mohammadi, M.Sc.
E-Mail: farid.mohammadi@iws.uni-stuttgart.de
Department of Hydromechanics and Modelling of Hydrosystems (LH2)
Institute for Modelling Hydraulic and Environmental Systems (IWS), University
of Stuttgart www.iws.uni-stuttgart.de/lh2/
Pfaffenwaldring 61
70569 Stuttgart

Created on Wed Jul 10 2019

"""
import os
import numpy as np
import pandas as pd
import joblib
from PyLink.PyLinkForwardModel import PyLinkForwardModel
from surrogate_models.Input import Input
from surrogate_models.surrogate_models import MetaModel
from PostProcessing.PostProcessing import PostProcessing
from BayesInference.BayesInference import BayesInference, Discrepancy


if __name__ == "__main__":

    # =====================================================
    # =============   COMPUTATIONAL MODEL  ================
    # =====================================================
    Model = PyLinkForwardModel()

    Model.link_type = 'PyLink'
    Model.name = 'Beam9points'
    Model.input_file = "SSBeam_Deflection.inp"
    Model.input_template = "SSBeam_Deflection.tpl.inp"

    Model.shell_command = "myBeam9points SSBeam_Deflection.inp"
    Model.exe_path = os.getcwd()
    Model.Output.parser = 'read_Beam_Deflection'
    Model.Output.names = ['Deflection [m]']
    Model.Output.file_names = ["SSBeam_Deflection.out"]

    # For Bayesian inversion
    Model.meas_file = 'MeasuredData.csv'
    Model.meas_file_valid = 'MeasuredData_Valid.csv'

    # For Checking with the MonteCarlo refrence
    Model.mc_ref_file = 'MCrefs_MeanStd.csv'

    # =====================================================
    # =========   PROBABILISTIC INPUT MODEL  ==============
    # =====================================================
    # Define the uncertain parameters with their mean and
    # standard deviation
    Inputs = Input()

    Inputs.addMarginals()
    Inputs.Marginals[0].Name = 'Beam width'
    Inputs.Marginals[0].DistType = 'lognormal'
    Inputs.Marginals[0].Parameters = [0.15, 0.0075]

    Inputs.addMarginals()
    Inputs.Marginals[1].Name = 'Beam height'
    Inputs.Marginals[1].DistType = 'lognormal'
    Inputs.Marginals[1].Parameters = [0.3, 0.015]

    Inputs.addMarginals()
    Inputs.Marginals[2].Name = 'Youngs modulus'
    Inputs.Marginals[2].DistType = 'lognormal'
    Inputs.Marginals[2].Parameters = [30000e+6, 4500e+6]

    Inputs.addMarginals()
    Inputs.Marginals[3].Name = 'Uniform load'
    Inputs.Marginals[3].DistType = 'lognormal'
    Inputs.Marginals[3].Parameters = [1e4, 2e3]

    # =====================================================
    # ==========  DEFINITION OF THE METAMODEL  ============
    # =====================================================
    MetaModelOpts = MetaModel(Inputs)

    # Select if you want to preserve the spatial/temporal depencencies
    # MetaModelOpts.DimRedMethod = 'PCA'

    # Select your metamodel method
    # 1) PCE (Polynomial Chaos Expansion) 2) GPE (Gaussian Process Emulator)
    MetaModelOpts.metaModel = 'PCE'

    # ------------------------------------------------
    # ------------- PCE Specification ----------------
    # ------------------------------------------------
    # Select the sparse least-square minimization method for
    # the PCE coefficients calculation:
    # 1)OLS: Ordinary Least Square  2)BRR: Bayesian Ridge Regression
    # 3)LARS: Least angle regression  4)ARD: Bayesian ARD Regression
    # 5)FastARD: Fast Bayesian ARD Regression
    MetaModelOpts.RegMethod = 'FastARD'

    # Specify the max degree to be compared by the adaptive algorithm:
    # The degree with the lowest Leave-One-Out cross-validation (LOO)
    # error (or the highest score=1-LOO)estimator is chosen as the final
    # metamodel.
    MetaModelOpts.MinPceDegree = 6
    MetaModelOpts.MaxPceDegree = 6

    # Print summary of the regression results
    # MetaModelOpts.DisplayFlag = True

    # ------------------------------------------------
    # ------ Experimental Design Configuration -------
    # ------------------------------------------------
    MetaModelOpts.addExpDesign()

    # One-shot (normal) or Sequential Adaptive (sequential) Design
    MetaModelOpts.ExpDesign.Method = 'normal'
    MetaModelOpts.ExpDesign.NrSamples = 20

    # Sampling methods
    # 1) random 2) latin_hypercube 3) sobol 4) halton 5) hammersley
    # 6) chebyshev(FT) 7) korobov 8) grid(FT) 9) nested_grid(FT) 10)user
    MetaModelOpts.ExpDesign.SamplingMethod = 'latin_hypercube'

    # Sequential experimental design (needed only for sequential ExpDesign)
    MetaModelOpts.ExpDesign.NrofNewSample = 2
    MetaModelOpts.ExpDesign.MaxNSamples = 50
    MetaModelOpts.ExpDesign.ModifiedLOOThreshold = 1e-6

    DiscrepancyOpts = Discrepancy('')
    DiscrepancyOpts.Type = 'Gaussian'
    sigma2s = np.square([1e-8, 1e-2, 1e-2, 1e-3, 4e-3, 5e-3,
                        1e-3, 1e-2, 1e-2, 1e-2, 1e-2, 1e-9])
    DiscrepancyOpts.Parameters = sigma2s
    MetaModelOpts.Discrepancy = DiscrepancyOpts

    # Plot the posterior snapshots for SeqDesign
    MetaModelOpts.ExpDesign.PostSnapshot = True
    MetaModelOpts.ExpDesign.stepSnapshot = 1
    MetaModelOpts.ExpDesign.MAP = (0.150064, 0.299698, 30.763206e9, 10.164872e2)
    MetaModelOpts.ExpDesign.parNames = ['Beam width', 'Beam height',
                                        'Youngs modulus', 'Uniform load']

    # ------------------------------------------------
    # ------- Sequential Design configuration --------
    # ------------------------------------------------
    # 1) 'None' 2) 'epsilon-decreasing'
    MetaModelOpts.ExpDesign.TradeOffScheme = 'None'
    # MetaModelOpts.ExpDesign.nReprications = 2
    # -------- Exploration ------
    # 1)'Voronoi' 2)'MC' 3)'LHS' 4)'dual annealing'
    MetaModelOpts.ExpDesign.ExploreMethod = 'Voronoi'

    # Use when 'dual annealing' chosen
    MetaModelOpts.ExpDesign.MaxFunItr = 200

    # Use when 'Voronoi' or 'MC' or 'LHS' chosen
    MetaModelOpts.ExpDesign.NCandidate = 1000
    MetaModelOpts.ExpDesign.NrofCandGroups = 4

    # -------- Exploitation ------
    # 1)'BayesOptDesign' 2)'VarOptDesign' 3)'alphabetic' 4)'Space-filling'
    MetaModelOpts.ExpDesign.ExploitMethod = 'VarOptDesign'

    # BayesOptDesign -> when data is available
    # 1)DKL (Kullback-Leibler Divergence) 2)DPP (D-Posterior-percision)
    # 3)APP (A-Posterior-percision)
    # MetaModelOpts.ExpDesign.UtilityFunction = 'DKL' #['DKL', 'DPP']

    # VarBasedOptDesign -> when data is not available
    # Only with Vornoi >>> 1)Entropy 2)EIGF, 3)ALM, 4)LOOCV
    MetaModelOpts.ExpDesign.UtilityFunction = 'Entropy'

    # alphabetic
    # 1)D-Opt (D-Optimality) 2)A-Opt (A-Optimality)
    # 3)K-Opt (K-Optimality)
    # MetaModelOpts.ExpDesign.UtilityFunction = 'D-Opt'

    # >>>>>>>>>>>>>>>>>>>>>> Build Surrogate <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
    PCEModel = MetaModelOpts.create_metamodel(Model)

    # =====================================================
    # =========  POST PROCESSING OF METAMODELS  ===========
    # =====================================================
    PostPCE = PostProcessing(PCEModel)

    # Compute the moments and compare with the Monte-Carlo reference
    PostPCE.plotMoments()

    # Plot the sobol indices
    if MetaModelOpts.metaModel != 'GPE':
        sobol_cell, total_sobol = PostPCE.sobolIndicesPCE()

    # Plot to check validation visually.
    PostPCE.validMetamodel(nValidSamples=3)
    import sys
    sys.exit()
    # =====================================================
    # ============  Bayesian calibration  =================
    # =====================================================
    BayesOpts = BayesInference(PCEModel)
    BayesOpts.Name = 'Calib'
    BayesOpts.emulator = True

    # Evaluation of surrogate model predictions
    BayesOpts.NrofSamples = 50000

    # Bootstrap for BME calulations
    BayesOpts.Bootstrap = True
    BayesOpts.BootstrapItrNr = 1

    BayesOpts.PlotPostDist = True
    BayesOpts.PlotPostPred = True

    # ----- Define the discrepancy model -------
    # (Option A)
    # Do nothing

    # (Option B)
    obsData = pd.read_csv(Model.MeasurementFile)
    DiscrepancyOpts = Discrepancy('')
    DiscrepancyOpts.Type = 'Gaussian'
    DiscrepancyOpts.Parameters = (0.01*obsData)**2
    BayesOpts.Discrepancy = DiscrepancyOpts

    # (Option C)
    # DiscInputOpts = Input()
    # DiscInputOpts.addMarginals()
    # DiscInputOpts.Marginals[0].Name = 'Sigma2'
    # DiscInputOpts.Marginals[0].DistType = 'unif'
    # DiscInputOpts.Marginals[0].Parameters =  [0, 5e-1]

    # DiscrepancyOpts = Discrepancy(DiscInputOpts)

    # BayesOpts.Discrepancy = DiscrepancyOpts

    Bayes_Calib = BayesOpts.create_Inference()

    # =====================================================
    # ======= Compare Posterior distributions =============
    # =====================================================
    def posteriorPlot(Posterior, MAP, parNames, key, figsize=(10, 10)):
        import corner
        newpath = (r'Outputs_PosteriorComparison')
        if not os.path.exists(newpath):
            os.makedirs(newpath)

        nParams = len(parNames)
        figPosterior = corner.corner(Posterior, labels=parNames,
                                     show_titles=True,
                                     title_kwargs={"fontsize": 12})

        # This is the true mean of the second mode that we used above:
        value1 = MAP

        # This is the empirical mean of the sample:
        value2 = np.mean(Posterior, axis=0)

        # Extract the axes
        axes = np.array(figPosterior.axes).reshape((nParams, nParams))

        # Loop over the diagonal
        for i in range(nParams):
            ax = axes[i, i]
            ax.axvline(value1[i], color="g")
            ax.axvline(value2[i], ls='--', color="r")

        # Loop over the histograms
        for yi in range(nParams):
            for xi in range(yi):
                ax = axes[yi, xi]
                ax.axvline(value1[xi], color="g")
                ax.axvline(value2[xi], ls='--', color="r")
                ax.axhline(value1[yi], color="g")
                ax.axhline(value2[yi], ls='--', color="r")
                ax.plot(value1[xi], value1[yi], "sg")
                ax.plot(value2[xi], value2[yi], "sr")

        figPosterior.set_size_inches(figsize)
        figPosterior.savefig(f'./{newpath}/{key}.svg', bbox_inches='tight')

    theta_true = (0.150064, 0.299698, 30.763206, 10.164872)
    parNames = ['Beam width', 'Beam height', 'Youngs modulus', 'Uniform load']
    # Extract the posterior with emulator
    surr_posterior = Bayes_Calib.Posterior_df

    # Plot surrogate posteriors
    posteriorPlot(surr_posterior, theta_true, parNames, 'surrogatePosterior')

    # Read the original posterior
    orig_posterior = pd.read_csv("origPosterior.csv")

    # Plot original posteriors
    posteriorPlot(orig_posterior, theta_true, parNames, 'origPosterior')

    # =====================================================
    # =============  Bayesian validation  =================
    # =====================================================
    BayesOpts_Valid = BayesInference(PCEModel)
    BayesOpts_Valid.Name = 'Valid'
    BayesOpts_Valid.Samples = Bayes_Calib.Posterior_df

    BayesOpts_Valid.emulator = True

    # Bootstrap for BME calulations
    BayesOpts_Valid.Bootstrap = True
    BayesOpts_Valid.BootstrapItrNr = 20

    BayesOpts_Valid.PlotPostDist = True
    BayesOpts_Valid.PlotPostPred = True

    # ----- Define the discrepancy model -------
    # (Option A)
    # Do nothing

    # (Option B)
    # DiscInputOpts_Valid = Discrepancy('')
    # DiscInputOpts_Valid.Type = 'Gaussian'
    # DiscInputOpts_Valid.Parameters = [0, 1e-3, 1e-3, 1e-2, 1e-2, 1e-2, 1e-2,
    #                                   1e-3, 1e-3, 1e-3, 1e-3, 0] #1e-5

    # (Option C)
    # DiscInputOpts_Valid = Input()

    # DiscInputOpts_Valid.addMarginals()
    # DiscInputOpts_Valid.Marginals[0].Name = 'Sigma2'
    # DiscInputOpts_Valid.Marginals[0].DistType = 'unif'
    # DiscInputOpts_Valid.Marginals[0].Parameters =  [0, 5e-3]

    # DiscrepancyOptsValid = Discrepancy(DiscInputOpts_Valid)

    # BayesOpts_Valid.Discrepancy = DiscrepancyOptsValid

    Bayes_Valid = BayesOpts_Valid.create_Inference()

    # =====================================================
    # ==============  Save class objects  =================
    # =====================================================
    with open('Beam_Results.pkl', 'wb') as output:
        joblib.dump(PCEModel, output, 2)
        joblib.dump(PostPCE, output, 2)
        joblib.dump(Bayes_Calib, output, 2)
        joblib.dump(Bayes_Valid, output, 2)
